/*
* util.h
* Luabins test utilities
* See copyright notice in luabins.h
*/

#ifndef LUABINS_TEST_UTIL_H_INCLUDED_
#define LUABINS_TEST_UTIL_H_INCLUDED_

#include <stdio.h>

void fprintbuf(FILE * out, const unsigned char * b, size_t len);

#endif /* LUABINS_TEST_UTIL_H_INCLUDED_ */
