// Copyright (c) 2013, Thomas Goyne <plorkyeran@aegisub.org>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
// Aegisub Project http://www.aegisub.org/

#include <libaegisub/fs.h>
#include <memory>
#include <string>

namespace agi {
	namespace charset { class IconvWrapper; }
	namespace io { class Save; }
}

class TextFileWriter {
	std::unique_ptr<agi::io::Save> file;
	std::unique_ptr<agi::charset::IconvWrapper> conv;
#ifdef _WIN32
	std::string newline = "\r\n";
#else
	std::string newline = "\n";
#endif

public:
	TextFileWriter(agi::fs::path const& filename, std::string encoding="");
	~TextFileWriter();

	void WriteLineToFile(std::string_view line, bool addLineBreak=true);
};
