// Copyright (c) 2014, Thomas Goyne <plorkyeran@aegisub.org>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
// Aegisub Project http://www.aegisub.org/

#include <libaegisub/vfr.h>

#include <string>

namespace agi {
class Time;

/// @class SmpteFormatter
/// @brief Convert times to and from SMPTE timecodes
class SmpteFormatter {
	/// Frame rate to use
	vfr::Framerate fps;
	/// Separator character
	char sep;

public:
	SmpteFormatter(vfr::Framerate fps, char sep=':');

	/// Convert an Time to a SMPTE timecode
	std::string ToSMPTE(Time time) const;
	/// Convert a SMPTE timecode to an Time
	Time FromSMPTE(std::string const& str) const;
};
}
